-- Simple spiral text demo

module SpiralText where

import Monad
import Pan

{-
Given a sized image, tile horizontally and wrap the infinite horizontal
strip in a spiral.  Positive x gets squeezed, going into the spiral and
negative x gets stretched.  Each full turn doubles or halves the scale.

See DerivSpiralText.{tex,dvi,ps,pdf} for derivation.  (Ok, it's on paper
for now.)
-}

spiralify :: BoolE -> FloatE -> FloatE -> SizedImage a -> FloatE -> Image a
spiralify tileH multH radiusPerH (w,h',im) motionPerW = tiledIm . f . toPolar
 where 
   f (rho,theta') = (x,y)
     where
       -- Normalize angle to [0,2pi)
       theta = --wrap (2*pi) theta  -- crashes hugs
               --ifE (theta' <* 0) (theta'+pi) theta'
               -- Hmm... this is the one that works.  Why?
               theta'
       thetaFrac = theta / (2*pi)

       thetaFull = theta + 2*pi * i2f n

       x = - thetaFull * radius
       y = rho * b**(-x/circum) - radius

       n = floorE (logBase b ((radius - halfH)/rho) - thetaFrac)

   h = h' * multH
   radius = radiusPerH * h
   circum = 2*pi * radius

   halfH = h/2
   -- Scaling per period (2pi)
   b = (radius - halfH)/(radius + halfH)

   tiledIm = translate (motionPerW*w,0) $
             im . ifE tileH (\ (x,y) -> (wrap' w x, y)) id


spiralSImUI :: UI (SizedImage a) -> UI (FloatE -> Image a)
spiralSImUI simUI = liftM4 spiralify
                      (tick "repeat" True)
                      (slider "multiply height" (-0.2,5.0) 1)
                      (slider "radius/height" (0.1,5) 1)
                      simUI

rateSlider :: String -> (Double,Double) -> Double -> UI (TimeE -> FloatE)
rateSlider label bounds dflt =
  liftM (*) (slider label bounds dflt)



motionUI = slider "motion/width" (-3,3) 0
rateUI   = rateSlider "rate/width" (-1,1) (-0.5)

spiralTextUI, spiralImgUI :: UI ImageC

spiralTextUI = spiralSImUI (textBoxSize " a spiral text message -")
               `ap` motionUI

spiralImgUI = spiralSImUI (browseImageSizeSpread "source" "flower.bmp")
               `ap` motionUI

spiralTextTurningUI, spiralImgTurningUI :: UI ImageBC

spiralTextTurningUI =
  spiralSImUI (textBoxSize " a spiral text message -")
   `comp` rateUI

spiralImgTurningUI =
  spiralSImUI (browseImageSizeSpread "source" "flower.bmp")
   `comp` rateUI


browseImageSizeSpread label fileName =
  do hsep     <- slider "horizontal spread" (0.2,2) 1
     (w,h,im) <- browseImageSize label fileName
     return (hsep*w,h,im)

{-
spiralSImUI :: UI (SizedImage a) -> UI (ImageB a)
spiralSImUI simUI = liftM5 spiralify
                      (tick "repeat" True)
                      (slider "multiply height" (-0.2,5.0) 1)
                      (slider "radius/height" (0.1,5) 1)
                      simUI
                      (slider "rate/height" (-10,10) (-2))

spiralTextUI :: UI ImageBC
spiralTextUI = spiralSImUI $
                 textBoxSize " a spiral text message -"

spiralImgUI :: UI ImageBC
spiralImgUI = spiralSImUI $
                 browseImageSizeSpread "source" "flower.bmp"

-- Equivalently:
browseImageSizeSpread' label fileName =
  liftM2 f (slider "horizontal spread" (0.2,2) 1)
           (browseImageSize label fileName)
 where
   f spread (w,h,im) = (spread*w,h,im)

  
-- Like toPolar, but adjusts angle from [-pi,pi) to [0,2*pi)
toPolar' p = (r, ifE (a <* 0) (a+pi) a)
 where
   (r,a) = toPolar p
-}

----------------------------------------

makeAll = makeAllFor allBackends

makeAllFor bends = do
  compileImageFor  bends "spiral text"  spiralTextUI
  compileImageFor  bends "spiral image" spiralImgUI
  compileImageBFor bends "spiral text turning"  spiralTextTurningUI
  compileImageBFor bends "spiral image turning" spiralImgTurningUI
