-- Misc definitions shared by demos, but too specialized to go into Pan
-- 
-- ToDo: move some of them into Pan, as appropriate

module DemoUtils where

import Monad
import Exp
import PointE
import Color
import TransformE
import Image
import DisplayConvert
import UI
import Compiler


-- Default view size for standalone
viewSize = 300 :: Int

---- Some images

cb0, cb1, bilerpBRBW, tiledBigBilerp :: ImageC

-- B&W checkerboard
cb0 = checkerBoard 10 blackI whiteI

-- Colorful checkerBoard
cb1 = uscale 20 $
      about (0.5,0.5) (tile (1,1)) $
      bilerpBRBW

bilerpBRBW     = bilerpC black red blue white
tiledBigBilerp = uscale 5 cb1


wavDist :: Image FractionE
wavDist p = (1 + cos (pi * distO p)) / 2

ybRings :: ImageC
ybRings = lerpI (uscale 15 wavDist) blueI yellowI

vstrip :: Region
vstrip (x,y) = abs x <=* 1/2


type SizedImage a = (FloatE,FloatE,Image a)
type SizedImageC  = SizedImage ColorE

---- UI tools

-- Move an Filter to a the mouse point.  [move into DemoUtils]
atMouse :: UI (Filter a) -> UI (Filter a)
atMouse xfUI = liftM2 about mouseInput xfUI

atMouseB :: UI (Image a -> ImageB a) -> UI (Image a -> ImageB a)
atMouseB filtBUI = liftM2 aboutB mouseInput filtBUI
 where
   aboutB pos filtB im t = about pos (flip filtB t) im

progressFromUI :: Fraction -> UI FractionE
progressFromUI = slider "progress" (0,1)

progressUI :: UI FractionE
progressUI = progressFromUI 0.5

-- Scale [0,1] to [0,2pi] for progress UI.
progressAngleUI :: UI TimeE
progressAngleUI = do p <- progressUI
                     return $ p * 2*pi


sizeUI :: UI (FloatE,FloatE)
sizeUI = do w <- slider "width"  (0,500) 100
            h <- slider "height" (0,500) 100
            return (w,h)

boxUI :: UI Region
boxUI = do (w,h) <- sizeUI 
           return $ box w h

-- The angleUI function used here is defined in terms of the slider
-- primitive.  It interprets the slider value as degrees, with positive to
-- the right (clockwise) and negative to the left
-- (counter-clockwise). This convention matches the slider,
-- kinesthetically, and is mapped to radians with positive
-- counter-clockwise and negative clockwise.
angleUI :: Label -> (Double, Double) -> Double -> UI FloatE
angleUI lab range def = do angDeg <- slider lab range def
                           return $ angDeg * (-pi/180)

fullAngleUI :: UI FloatE
fullAngleUI = angleUI "angle" (-180,180) 0

powerZoomUI :: Label -> Double -> Double -> UI (Filter a)
powerZoomUI label max dflt =
  do zoom <- slider label (-max,max) dflt
     return $ uscale (2 ** zoom)

-- PhS filters now do mouse-based pan&zoom
-- navUI :: UI (Image a -> Image a)
-- navUI = powerZoomUI "power zoom" 5 0


-- Time-varying combiner of two images
type ImageCombiner = ImageC -> ImageC -> ImageBC


---- Conversions

class ToColor a where toColor :: a -> ColorE

instance ToColor ColorE    where toColor = id
instance ToColor BoolE     where toColor = \ b -> ifE b black invisible
instance ToColor FractionE where toColor = flip fadeC black

toImageC :: ToColor a => Image a -> ImageC
toImageC = lift1 toColor


---- Compiler functions.

---- Convenient compiler entry points

allBackends :: [Backend]
allBackends = [Standalone,Component,Control,DXTransform,PhotoShopFilter]

-- At some point, we may discontinue Standalone.

-- All types are compiled by converting to FilterBC

type Compiler    a = Label -> UI a -> IO ()
type CompilerFor a = [Backend] -> Compiler a

compileWith :: (Backend -> Compiler a) -> CompilerFor a
compileWith f ends nm ui = sequence_ (map (\ end -> f end nm ui) ends)


compileImageFor :: ToColor c => CompilerFor (Image c)
compileImageFor = compileWith compileImageBack

compileImage :: ToColor c => Compiler (Image c)
compileImage = compileImageFor allBackends

compileImageBFor :: ToColor c => CompilerFor (ImageB c)
compileImageBFor = compileWith compileImageBBack

compileImageB :: ToColor c => Compiler (ImageB c)
compileImageB = compileImageBFor allBackends

compileFilterFor :: ToColor c => CompilerFor (Filter' c)
compileFilterFor = compileWith compileFilterBack

compileFilter :: ToColor c => Compiler (Filter' c)
compileFilter = compileFilterFor allBackends

compileFilterBFor :: ToColor c => CompilerFor (FilterB' c)
compileFilterBFor = compileWith compileFilterBBack

compileFilterB :: ToColor c => Compiler (FilterB' c)
compileFilterB = compileFilterBFor allBackends

compileCombinerFor :: CompilerFor ImageCombiner
compileCombinerFor = compileWith compileCombinerBack

compileCombiner :: Compiler ImageCombiner
compileCombiner = compileCombinerFor allBackends


---- The building blocks.  These ones compile for a specified backend.

compileImageBack :: ToColor c => Backend -> Label -> UI (Image c) -> IO ()
compileImageBack end nm ui = compileImageBBack end nm (liftM const ui)

compileImageBBack :: ToColor c => Backend -> Label -> UI (ImageB c) -> IO ()

-- For plugin, we want to scale automatically to account for difference
-- between preview and final version.  The same happens for filters, so
-- convert to a filter
compileImageBBack PhotoShopFilter nm ui =
  compileFilterBBack PhotoShopFilter nm $
  liftM (\ imb -> \ src t -> imb t) ui

compileImageBBack DXTransform nm ui =
  compileFilterBBack DXTransform nm $
  liftM (\ imb -> \ src t -> imb t) ui

-- Otherwise, wrap a toImageC and do simple display
compileImageBBack end nm imbUI = compileDispUI end nm (liftM (toImageC .) imbUI)


-- I don't think these types are *really* necessary.  compileFilter could
-- take a Filter c and compileImage could take an Image c, except for how
-- I use compileFilter in defining compileImage.
type Filter'  c = ImageC -> Image c
type FilterB' c = ImageC -> ImageB c

compileFilterBack :: ToColor c => Backend -> Label -> UI (Filter' c) -> IO ()
compileFilterBack end nm filterUI = compileFilterBBack end nm $
  liftM (\ filter -> \ src t -> filter src) filterUI

compileFilterBBack :: ToColor c => Backend -> Label -> UI (FilterB' c) -> IO ()
compileFilterBBack end nm filtBui = compileDispUI  end nm $
  do navF <- return id
             -- PhS filters now do mouse-based pan&zoom
             -- if end == PhotoShopFilter then navUI else return id
     (w,h,im) <- imageSourceSize "source image"
     imF      <- filtBui
     let combine = max   -- \ w h -> (w+h)/2
         sc      = 100 / combine w h
     return $ \ t -> navF (toImageC (uscale (1/sc) (imF (uscale sc im) t)))

compileCombinerBack :: Backend -> Label -> UI ImageCombiner -> IO ()
compileCombinerBack end nm combUI =
  compileFilterBBack end nm $
  do combine <- combUI
     imB     <- imageSource "other image"
     return (`combine` imB)

compileDispUI :: Displayable a => Backend -> Label -> UI a -> IO ()
compileDispUI backend nm imUI =
  genUI (Just pluginInfo{name=nm}) False backend
        (wrap (imUI >>= display (viewSize,viewSize)))
   where
     wrap | backend == PhotoShopFilter = okCancel
          | otherwise                  = id
     display = case backend of
                 Standalone      -> mkDisplay
                 Component       -> mkDisplay
		 Control         -> mkDisplay
                 DXTransform     -> dxtDisplay
                 PhotoShopFilter -> pluginDisplay




--------------- Misc

-- Like ap, but "does" arg before function.
ap' :: Monad m => m (a->b) -> m a -> m b
ap' fui argui = do arg <- argui
                   f   <- fui
                   return (f arg)

comp :: Monad m => m (b -> c) -> m (a -> b) -> m (a -> c)
comp = liftM2 (.)

comps :: Monad m => [m (a->a)] -> m (a->a)
comps = foldr comp (return id)


